"""Module for license server interface abstract base class."""
import abc
import typing

from pydantic import BaseModel, Field

from lm_agent.config import PRODUCT_FEATURE_RX


class LicenseServerInterface(metaclass=abc.ABCMeta):
    """
    Abstract base class for License Server interface.

    The logic for obtaining the data output from the License Server should be encapsulated in
    the ``get_output_from_server`` method.

    After obtaining the output, the parsing and manipulation of the data should be implemented in
    the ``get_report_item`` method.

    The license information should be parsed into a ``LicenseReportItem``.
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            all(
                [
                    hasattr(subclass, "get_output_from_server"),
                    callable(subclass.get_output_from_server),
                    hasattr(subclass, "get_report_item"),
                    callable(subclass.get_report_item),
                ]
            )
            or NotImplemented
        )

    @abc.abstractclassmethod
    def get_output_from_server(self, product_feature: str):
        """Return output from license server for the indicated features."""

    @abc.abstractclassmethod
    def get_report_item(self, product_feature: str):
        """Parse license server output into a report item for the indicated feature."""


class LicenseReportItem(BaseModel):
    """
    An item in a LicenseReport, a count of tokens for one product/feature.
    """

    product_feature: str = Field(..., regex=PRODUCT_FEATURE_RX)
    used: int
    total: int
    used_licenses: typing.List



#!/usr/bin/env python3
"""
The SlurmctldProlog executable.

This prolog is responsible for checking if feature tokens are available
and making booking requests by communicating with the license-manager agent
(which should be running on the slurmctld host).

Executing this script will result in either an exit(0) or exit(1). Slurm will
proceed with scheduling the job if the exit status is 0, and will not proceed
if the exit status is anything other then 0, e.g. 1.
"""
import asyncio
import sys

from lm_agent.backend_utils import get_config_from_backend
from lm_agent.config import settings
from lm_agent.logs import init_logging, logger
from lm_agent.reconciliation import update_report
from lm_agent.workload_managers.slurm.cmd_utils import (
    LicenseBookingRequest,
    get_required_licenses_for_job,
    make_booking_request,
)
from lm_agent.workload_managers.slurm.common import get_job_context


async def prolog():
    """The PrologSlurmctld for the license-manager-agent."""
    # Initialize the logger
    init_logging("slurmctld-prolog")
    # Acqure the job context
    job_context = get_job_context()
    job_id = job_context.get("job_id", "")
    user_name = job_context.get("user_name")
    lead_host = job_context.get("lead_host")
    cluster_name = job_context.get("cluster_name")
    job_licenses = job_context.get("job_licenses")

    logger.info(f"Prolog started for job id: {job_id}")

    try:
        required_licenses = get_required_licenses_for_job(job_licenses)
        logger.debug(f"Required licenses: {required_licenses}")
    except Exception as e:
        logger.error(f"Failed to call get_required_licenses_for_job with {e}")
        sys.exit(1)

    if not required_licenses:
        logger.debug("No licenses required, exiting!")
        sys.exit(0)

    tracked_licenses = list()
    # Create a list of tracked licenses in the form <product>.<feature>

    if len(required_licenses) > 0:
        # Create a list of tracked licenses in the form <product>.<feature>
        try:
            entries = await get_config_from_backend()
        except Exception as e:
            logger.error(f"Failed to call get_config_from_backend with {e}")
            sys.exit(1)
        for entry in entries:
            for feature in entry.features:
                tracked_licenses.append(f"{entry.product}.{feature}")
    logger.debug(f"Tracked licenses: {tracked_licenses}")

    # Create a tracked LicenseBookingRequest for licenses that we actually
    # track. These tracked licenses are what we will check feature token
    # availability for.
    tracked_license_booking_request = LicenseBookingRequest(
        job_id=job_id,
        bookings=[],
        user_name=user_name,
        lead_host=lead_host,
        cluster_name=cluster_name,
    )
    for booking in required_licenses:
        if booking.product_feature in tracked_licenses:
            tracked_license_booking_request.bookings.append(booking)
    logger.debug(f"Tracked license bookings: {tracked_license_booking_request}")

    if len(tracked_license_booking_request.bookings) > 0:
        # Check if reconciliation should be triggered.
        if settings.USE_RECONCILE_IN_PROLOG_EPILOG:
            # Force a reconciliation before we check the feature token availability.
            try:
                await update_report()
            except Exception as e:
                logger.error(f"Failed to call reconcile with {e}")
                sys.exit(1)

        booking_request = await make_booking_request(tracked_license_booking_request)
        if not booking_request:
            logger.debug("Booking request unsuccessful, not enough licenses.")
            sys.exit(1)
        logger.debug(f"License booking sucessful, job id: {job_id}.")
        logger.debug(f"Licenses booked: {repr(tracked_license_booking_request.bookings)}")
    sys.exit(0)


def main():
    asyncio.run(prolog())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
The EpilogSlurmctld executable.
"""
import asyncio
import sys

from lm_agent.backend_utils import backend_client
from lm_agent.config import settings
from lm_agent.logs import init_logging, logger
from lm_agent.reconciliation import update_report
from lm_agent.workload_managers.slurm.cmd_utils import get_required_licenses_for_job
from lm_agent.workload_managers.slurm.common import get_job_context


async def _remove_booking_for_job(job_id: str) -> bool:
    """Remove token bookings used by job."""

    # Remove the booking for the job.
    resp = await backend_client.delete(f"lm/api/v1/booking/book/{job_id}")
    # Return True if the request to delete the booking was successful.
    if resp.status_code == 200:
        return True
    return False


async def epilog():
    # Initialize the logger
    init_logging("slurmctld-epilog")
    job_context = get_job_context()
    job_id = job_context["job_id"]
    job_licenses = job_context["job_licenses"]

    # Check if reconciliation should be triggered.
    if settings.USE_RECONCILE_IN_PROLOG_EPILOG:
        # Force a reconciliation before we attempt to remove bookings.
        try:
            await update_report()
        except Exception as e:
            logger.error(f"Failed to call reconcile with {e}")
            sys.exit(1)

    try:
        required_licenses = get_required_licenses_for_job(job_licenses)
    except Exception as e:
        logger.error(f"Failed to call get_required_licenses_for_job with {e}")
        sys.exit(1)

    if not required_licenses:
        logger.debug("No licenses required, exiting!")
        sys.exit(0)

    if len(required_licenses) > 0:
        # Attempt to remove the booking and log the result.
        booking_removed = await _remove_booking_for_job(job_id)
        if booking_removed:
            logger.debug(f"Booking for job id: {job_id} successfully deleted.")
        else:
            logger.debug(f"Booking for job id: {job_id} not removed.")


def main():
    asyncio.run(epilog())


if __name__ == "__main__":
    main()



"""
Provide utilities that communicate with the backend.
"""
import shutil
import typing

import httpx
import jwt
from pydantic import BaseModel, ValidationError

from lm_agent.config import settings
from lm_agent.exceptions import LicenseManagerAuthTokenError, LicenseManagerBackendConnectionError
from lm_agent.logs import logger

TOKEN_FILE_NAME = "access.token"


def _load_token_from_cache() -> typing.Union[str, None]:
    """
    Looks for and returns a token from a cache file (if it exists).

    Returns None if::
    * The token does not exist
    * Can't read the token
    * The token is expired (or will expire within 10 seconds)
    """
    token_path = settings.CACHE_DIR / TOKEN_FILE_NAME
    if not token_path.exists():
        return None

    try:
        token = token_path.read_text()
    except Exception as err:
        logger.warning(f"Couldn't load token from cache file {token_path}. Will acquire a new one: {err}")
        return None

    try:
        jwt.decode(token, options=dict(verify_signature=False, verify_exp=True), leeway=-10)
    except jwt.ExpiredSignatureError:
        logger.warning("Cached token is expired. Will acquire a new one.")
        return None

    return token


def _write_token_to_cache(token: str):
    """
    Writes the token to the cache.
    """
    cache_dir = settings.CACHE_DIR
    if not cache_dir.exists():
        logger.warning(f"Cache directory does not exist {cache_dir}. Token won't be saved.")
        return

    token_path = settings.CACHE_DIR / TOKEN_FILE_NAME
    try:
        token_path.touch(mode=0o600, exist_ok=True)
        token_path.write_text(token)
        shutil.chown(token_path, "slurm", "slurm")
    except Exception as err:
        logger.warning(f"Couldn't save token to {token_path}: {err}")


def acquire_token() -> str:
    """
    Retrieves a token from OIDC based on the app settings.
    """
    logger.debug("Attempting to use cached token")
    token = _load_token_from_cache()

    if token is None:
        logger.debug("Attempting to acquire token from OIDC")
        oidc_body = dict(
            audience=settings.OIDC_AUDIENCE,
            client_id=settings.OIDC_CLIENT_ID,
            client_secret=settings.OIDC_CLIENT_SECRET,
            grant_type="client_credentials",
        )
        oidc_url = f"https://{settings.OIDC_DOMAIN}/protocol/openid-connect/token"
        logger.debug(f"Posting OIDC request to {oidc_url}")
        response = httpx.post(oidc_url, data=oidc_body)
        LicenseManagerAuthTokenError.require_condition(
            response.status_code == 200, f"Failed to get auth token from OIDC: {response.text}"
        )
        with LicenseManagerAuthTokenError.handle_errors("Malformed response payload from OIDC"):
            token = response.json()["access_token"]
        _write_token_to_cache(token)

    logger.debug("Successfully acquired auth token from OIDC")
    return token


class AsyncBackendClient(httpx.AsyncClient):
    """
    Extends the httpx.AsyncClient class with automatic token acquisition for requests.
    The token is acquired lazily on the first httpx request issued.

    This client should be used for most agent actions.
    """

    _token: typing.Optional[str]

    def __init__(self):
        self._token = None
        super().__init__(base_url=settings.BACKEND_BASE_URL, auth=self._inject_token)

    def _inject_token(self, request: httpx.Request) -> httpx.Request:
        if self._token is None:
            self._token = acquire_token()
        request.headers["authorization"] = f"Bearer {self._token}"
        return request


backend_client = AsyncBackendClient()


class SyncBackendClient(httpx.Client):
    """
    Extends the synchronous httpx.Client class with automatic token acquisition for requests.
    The token is acquired lazily on the first httpx request issued.

    This client should be used only for the CLI tools.
    """

    _token: typing.Optional[str]

    def __init__(self):
        self._token = None
        super().__init__(base_url=settings.BACKEND_BASE_URL, auth=self._inject_token)

    def _inject_token(self, request: httpx.Request) -> httpx.Request:
        if self._token is None:
            self._token = acquire_token()
        request.headers["authorization"] = f"Bearer {self._token}"
        return request


async def check_backend_health():
    """Hit the API's health-check endpoint to make sure the API is available."""
    resp = await backend_client.get("/lm/health")
    if resp.status_code != 204:
        logger.error("license-manager-backend health-check failed.")
        raise LicenseManagerBackendConnectionError("Could not connect to the backend health-check endpoint")


class BackendConfigurationRow(BaseModel):
    """
    NOTE: This is a copy of the schema from the backend.
          If the schema changes upstream in a non-reverse-compatible way, this schema should cause errors
          in deserialization.
    """

    class Config:
        extra = "ignore"

    id: typing.Optional[int] = None
    product: str
    features: dict
    license_servers: typing.List[str]
    license_server_type: str
    grace_time: int
    client_id: str


class BackendBookingRow(BaseModel):
    """
    NOTE: This is a copy of the schema from the backend.
          If the schema changes upstream in a non-reverse-compatible
          way, this schema should cause errors in deserialization.
    """

    class Config:
        extra = "ignore"

    job_id: str
    product_feature: str
    booked: int
    config_id: int
    lead_host: str
    user_name: str
    cluster_name: str


async def get_bookings_from_backend(
    cluster_name: typing.Optional[str] = None,
) -> typing.List[BackendBookingRow]:
    bookings: typing.List = []
    try:
        if cluster_name:
            resp = await backend_client.get(f"/lm/api/v1/booking/all?cluster_name={cluster_name}")
        else:
            resp = await backend_client.get("/lm/api/v1/booking/all")
    except httpx.ConnectError as e:
        logger.error(f"Connection failed to backend: {e}")
        return bookings
    for booking in resp.json():
        try:
            bookings.append(BackendBookingRow.parse_obj(booking))
        except ValidationError as err:
            logger.error(f"Wrong format for booking: {str(err)}")
    return bookings


async def get_config_id_from_backend(product_feature: str) -> int:
    """Given the product_feature return return the config id."""
    path = "/lm/api/v1/config/"
    resp = await backend_client.get(path, params={"product_feature": product_feature})
    return resp.json()


async def get_config_from_backend() -> typing.List[BackendConfigurationRow]:
    """Get all config rows from the backend."""
    path = "/lm/api/v1/config/agent/all"

    try:
        resp = await backend_client.get(path)
    except httpx.ConnectError as e:
        logger.error(f"Connection failed to backend: {backend_client.base_url}{path}: {e}")
        return []

    configs = []
    for (i, config_row) in enumerate(resp.json()):
        try:
            configs.append(BackendConfigurationRow.parse_obj(config_row))
        except ValidationError as err:
            logger.error(f"Could not validate config entry at row {i}: {str(err)}")
    return configs



"""Utilities that interact with slurm."""
import asyncio
import re
import shlex
import subprocess
from typing import List, Optional, Union

from pydantic import BaseModel, Field

from lm_agent.backend_utils import backend_client
from lm_agent.config import PRODUCT_FEATURE_RX
from lm_agent.logs import logger
from lm_agent.workload_managers.slurm.common import (
    CMD_TIMEOUT,
    ENCODING,
    SACCTMGR_PATH,
    SCONTROL_PATH,
    SQUEUE_PATH,
)


class SqueueParserUnexpectedInputError(ValueError):
    """Unexpected squeue output."""


class ScontrolRetrievalFailure(Exception):
    """
    Could not get SLURM data for job id.
    The following function's return code was zero:

    await asyncio.create_subprocess_shell(
        shlex.join([SCONTROL_PATH, "show", f"job={slurm_job_id}"]),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    """


class LicenseBooking(BaseModel):
    """
    Structure to represent a license booking.
    """

    product_feature: str = Field(..., regex=PRODUCT_FEATURE_RX)
    tokens: int
    license_server_type: Union[None, str]


class LicenseBookingRequest(BaseModel):
    """
    Structure to represent a list of license bookings.
    """

    job_id: int
    bookings: Union[List, List[LicenseBooking]]
    user_name: str
    lead_host: str
    cluster_name: str


def _match_requested_license(requested_license: str) -> Union[dict, None]:
    license_regex = re.compile(r"(?P<product>\w+)\.(?P<feature>\w+)@(?P<server_type>\w+)(:(?P<tokens>\d+))?")

    matches = license_regex.match(requested_license)

    if not matches:
        return None

    groups = matches.groupdict()
    if not groups["tokens"]:
        groups["tokens"] = "1"

    return {
        "product_feature": groups["product"] + "." + groups["feature"],
        "server_type": groups["server_type"],
        "tokens": int(groups["tokens"]),
    }


def get_required_licenses_for_job(job_licenses: str) -> List:
    """Retrieve the required licenses for a job."""

    license_array = job_licenses.split(",")
    logger.debug(f"##### License array for job: {license_array} #####")

    required_licenses: List = []

    if not license_array:
        return required_licenses

    if license_array[0] == "(null)":
        return required_licenses

    for requested_license in license_array:
        matched_license_items = _match_requested_license(requested_license)
        if not matched_license_items:
            continue
        product_feature = matched_license_items["product_feature"]
        license_server_type = matched_license_items["server_type"]
        tokens = matched_license_items["tokens"]
        # Create the license booking
        license_booking = LicenseBooking(
            product_feature=product_feature,
            tokens=tokens,
            license_server_type=license_server_type,
        )
        required_licenses.append(license_booking)

    return required_licenses


async def make_booking_request(lbr: LicenseBookingRequest) -> bool:
    """Book the feature tokens."""

    features = [
        {
            "product_feature": license_booking.product_feature,
            "booked": license_booking.tokens,
        }
        for license_booking in lbr.bookings
    ]

    logger.debug(f"features: {features}")
    logger.debug(f"lbr: {lbr}")

    resp = await backend_client.put(
        "/lm/api/v1/booking/book",
        json={
            "job_id": lbr.job_id,
            "features": features,
            "user_name": lbr.user_name,
            "lead_host": lbr.lead_host,
            "cluster_name": lbr.cluster_name,
        },
    )

    if resp.status_code == 200:
        logger.debug("##### Booking completed successfully #####")
        return True
    logger.debug(f"##### Booking failed: {str(resp.content)} #####")
    return False


async def reconcile():
    """Force a reconciliation."""
    resp = await backend_client.get("/lm/api/v1/license/reconcile")

    if resp.status_code == 200:
        logger.debug("##### Reconcile completed successfully #####")
        return True
    logger.debug(f"##### Reconcile failed {resp.status_code} #####")
    return False


async def get_tokens_for_license(
    product_feature_server: str,
    output_type: str,
) -> Optional[int]:
    """
    Return tokens from scontrol output.
    """

    def match_product_feature_server() -> Optional[str]:
        """
        Return the line after the matched product_feature line.
        """
        matched = False
        for line in scontrol_output.split("\n"):
            if matched:
                return line
            if len(re.findall(rf"({product_feature_server})", line)) > 0:
                matched = True
        return None

    # Get the scontrol output
    scontrol_output = await scontrol_show_lic()

    # Match the product_feature_server
    token_str = match_product_feature_server()
    if token_str is not None:
        for item in token_str.split():
            k, v = item.split("=")
            if k == output_type:
                return int(v)
    return None


async def scontrol_show_lic():
    """
    Get the license usage from scontrol.
    """

    cmd = [
        SCONTROL_PATH,
        "show",
        "lic",
    ]

    proc = await asyncio.create_subprocess_shell(
        shlex.join(cmd), stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT
    )

    stdout, _ = await asyncio.wait_for(proc.communicate(), CMD_TIMEOUT)
    output = str(stdout, encoding=ENCODING)
    logger.debug("##### scontrol show lic #####")
    logger.debug(output)
    return output


async def get_cluster_name() -> str:
    cmd = [
        SACCTMGR_PATH,
        "list",
        "cluster",
        "-nP",
        "format=Cluster",
    ]
    logger.debug("##### sacctmgr get cluster name cmd #####")
    logger.debug(f"{' '.join(cmd)}")

    sacctmgr_modify_resource = await asyncio.create_subprocess_shell(
        shlex.join(cmd),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    cluster_name_stdout, _ = await asyncio.wait_for(
        sacctmgr_modify_resource.communicate(),
        CMD_TIMEOUT,
    )
    return cluster_name_stdout.decode("utf8").strip()


async def sacctmgr_modify_resource(product: str, feature: str, num_tokens) -> bool:
    """
    Update the license resource in slurm.
    """
    cmd = [
        SACCTMGR_PATH,
        "modify",
        "resource",
        f"name={product}.{feature}",
        "set",
        f"count={num_tokens}",
        "-i",
    ]
    logger.debug("##### sacctmgr update cmd #####")
    logger.debug(f"{' '.join(cmd)}")

    sacctmgr_modify_resource = await asyncio.create_subprocess_shell(
        shlex.join(cmd),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    modify_resource_stdout, _ = await asyncio.wait_for(
        sacctmgr_modify_resource.communicate(),
        CMD_TIMEOUT,
    )

    rc = sacctmgr_modify_resource.returncode

    if not rc == 0:
        logger.warning(f"rc = {rc}!")
        logger.warning(modify_resource_stdout)
        return False
    return True


async def get_all_product_features_from_cluster() -> List[str]:
    """
    Returns a list of all product.feature in the cluster.
    """
    show_lic_output = await scontrol_show_lic()

    PRODUCT_FEATURE = r"LicenseName=(?P<product>[a-zA-Z0-9_]+)[_\-.](?P<feature>\w+)"
    RX_PRODUCT_FEATURE = re.compile(PRODUCT_FEATURE)

    parsed_features = []
    output = show_lic_output.split("\n")
    for line in output:
        parsed_line = RX_PRODUCT_FEATURE.match(line)
        if parsed_line:
            parsed_data = parsed_line.groupdict()
            product = parsed_data["product"]
            feature = parsed_data["feature"]
            parsed_features.append(f"{product}.{feature}")

    return parsed_features


def return_formatted_squeue_out() -> str:
    """
    Call squeue via Popen and return the formatted output.

    Return the squeue output in the form "<job_id>|<run_time>|<state>".
    """

    result = subprocess.run(
        f"{shlex.quote(SQUEUE_PATH)} --noheader --format='%A|%M|%T'",
        capture_output=True,
        shell=True,
        encoding="utf-8",
    )
    if result.returncode != 0:
        logger.error(result.stderr)
        raise Exception(result.stderr)

    return result.stdout.strip()


def _total_time_in_seconds(time_string: str) -> int:
    """
    Return the runtime in seconds for a job.

    This function takes a slurm time string ("<days>-<hours>:<minutes>:<seconds>") as input, parses
    and converts each of the units in the time string to seconds and returns a computed value, the sum of
    the days, hours, minutes and seconds (in seconds).
    """
    MINUTE = 60
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR

    days = 0
    hours = 0
    splitted_time = [int(value) for value in re.split("-|:", time_string)]
    splitted_time_len = len(splitted_time)

    if splitted_time_len == 4:
        days, hours, minutes, seconds = splitted_time
    elif splitted_time_len == 3:
        hours, minutes, seconds = splitted_time
    else:
        minutes, seconds = splitted_time

    return days * DAY + hours * HOUR + minutes * MINUTE + seconds


def squeue_parser(squeue_formatted_output) -> List:
    """Parse the squeue formatted output."""

    squeue_parsed_output: List = []

    if not squeue_formatted_output:
        return squeue_parsed_output

    def parse_squeue_line():
        """Parse a line from squeue formatted output."""
        try:
            job_id, run_time, state = line.split("|")
        except ValueError as e:
            logger.error(e)
            raise SqueueParserUnexpectedInputError()
        return job_id, run_time, state

    for line in squeue_formatted_output.split():
        job_id, run_time, state = parse_squeue_line()
        squeue_parsed_output.append(
            {
                "job_id": int(job_id),
                "run_time_in_seconds": _total_time_in_seconds(run_time),
                "state": state,
            }
        )

    return squeue_parsed_output



import bz2
import logging
import os
from logging.handlers import RotatingFileHandler

from lm_agent.config import settings

DEFAULT_FORMAT = "[%(asctime)s;%(levelname)s] %(filename)s:%(lineno)s - " "%(funcName)20s %(message)s"

level = getattr(logging, settings.LOG_LEVEL)
logger = logging.getLogger("license-manager-agent-logger")
logger.setLevel(level)


def _rotator(source: str, dest: str):
    """Rotate log and compress with bz2."""
    with open(source, "rb") as source_file:
        with bz2.open(dest, "wb") as bz2_file:
            bz2_file.write(source_file.read())
    os.remove(source)


def _namer(name: str):
    """Rename for the rotated log file."""
    return f"{name}.bz2"


def _formatter():
    """logging.Formatter configured the way we like it to look."""
    ret = logging.Formatter(DEFAULT_FORMAT)
    return ret


def init_logging(command_type: str):
    """Create and return the logging handler."""
    if settings.LOG_BASE_DIR is not None:
        _rotating_file_handler(f"{settings.LOG_BASE_DIR}/{command_type}.log")
    else:
        _default_logging()


def _rotating_file_handler(log_file: str):
    """Configure the rotating file handler."""
    global logger

    # Rotate the log file at 5MB, keep 10 rotations.
    handler = RotatingFileHandler(
        log_file,
        backupCount=10,
        maxBytes=5242880,
    )
    handler.rotator = _rotator
    handler.namer = _namer

    level = getattr(logging, settings.LOG_LEVEL)
    handler.setLevel(level)
    handler.setFormatter(_formatter())
    logger.addHandler(handler)


def _default_logging():
    """Create a default log handler.
    Create a default log handler so we can execute code that uses the logger
    outside of or before the logger is initialized in the entrypoint.
    """
    global logger

    handler = logging.StreamHandler()
    handler.setFormatter(_formatter())

    level = getattr(logging, settings.LOG_LEVEL)
    handler.setLevel(level)
    logger.addHandler(handler)


log = logger



"""RLM license server interface."""
import typing

from lm_agent.config import settings
from lm_agent.exceptions import LicenseManagerBadServerOutput
from lm_agent.parsing import rlm
from lm_agent.server_interfaces.license_server_interface import LicenseReportItem, LicenseServerInterface
from lm_agent.utils import run_command


class RLMLicenseServer(LicenseServerInterface):
    """Extract license information from RLM license server."""

    def __init__(self, license_servers: typing.List[str]):
        self.license_servers = license_servers
        self.parser = rlm.parse

    def get_commands_list(self):
        """Generate a list of commands with the available license server hosts."""

        host_ports = [(server.split(":")[1:]) for server in self.license_servers]
        commands_to_run = []
        for host, port in host_ports:
            command_line = f"{settings.RLMUTIL_PATH} rlmstat -c {port}@{host} -a -p"
            commands_to_run.append(command_line)
        return commands_to_run

    async def get_output_from_server(self):
        """Override abstract method to get output from RLM license server."""

        # get the list of commands for each license server host
        commands_to_run = self.get_commands_list()

        # run each command in the list, one at a time, until one succeds
        for cmd in commands_to_run:
            output = await run_command(cmd)

            # try the next server if the previous didn't return the expected data
            if not output:
                continue
            return output

        raise RuntimeError("None of the checks for RLM succeeded!")

    async def get_report_item(self, product_feature: str):
        """Override abstract method to parse RLM license server output into License Report Item."""

        server_output = await self.get_output_from_server()
        parsed_output = self.parser(server_output)

        (_, feature) = product_feature.split(".")

        current_feature_item = self._filter_current_feature(parsed_output["total"], feature)
        used_licenses = self._filter_used_features(parsed_output["uses"], feature)

        # raise exception if parser didn't output license information
        if current_feature_item is None or used_licenses is None:
            raise LicenseManagerBadServerOutput("Invalid data returned from parser.")

        report_item = LicenseReportItem(
            product_feature=product_feature,
            used=current_feature_item["used"],
            total=current_feature_item["total"],
            used_licenses=used_licenses,
        )

        return report_item

    def _filter_current_feature(self, parsed_list, feature):
        """
        The output from the RLM server returns information about all the licenses
        in the server. This function filters the output to return only the information
        about the feature we want.
        """
        for feature_item in parsed_list:
            if feature_item["feature"] == feature:
                return feature_item

    def _filter_used_features(self, parsed_list, feature):
        """
        The output from the RLM server returns information about all the licenses
        that are in use. This function filters the output to return only the information
        about the usage of the feature we want.
        """
        used_licenses = []
        for feature_booked in parsed_list:
            if feature_booked["feature"] == feature:
                used_licenses.append(feature_booked)

        for license in used_licenses:
            # remove the feature key, since we already handled it.
            del license["feature"]

        return used_licenses



#!/usr/bin/env python3
"""
Reconciliation functionality live here.
"""
import asyncio
from typing import Dict, List

from httpx import ConnectError

from lm_agent.backend_utils import (
    backend_client,
    get_bookings_from_backend,
    get_config_from_backend,
    get_config_id_from_backend,
)
from lm_agent.exceptions import (
    LicenseManagerBackendConnectionError,
    LicenseManagerEmptyReportError,
    LicenseManagerFeatureConfigurationIncorrect,
)
from lm_agent.logs import logger
from lm_agent.tokenstat import report
from lm_agent.workload_managers.slurm.cmd_utils import (
    get_all_product_features_from_cluster,
    get_cluster_name,
    get_tokens_for_license,
    return_formatted_squeue_out,
    sacctmgr_modify_resource,
    squeue_parser,
)

RECONCILE_URL_PATH = "/lm/api/v1/license/reconcile"


async def remove_booked_for_job_id(job_id: str):
    """
    Send DELETE to /lm/api/v1/booking/book/{job_id}.
    """
    response = await backend_client.delete(f"/lm/api/v1/booking/book/{job_id}")
    if response.status_code != 200:
        logger.error(f"{job_id} could not be deleted.")
        logger.debug(f"response from delete: {response.__dict__}")


async def get_all_grace_times() -> Dict[int, int]:
    """
    Send GET to /lm/api/v1/config/all.
    """
    response = await backend_client.get("/lm/api/v1/config/all")
    configs = response.json()
    grace_times = {config["id"]: config["grace_time"] for config in configs}
    return grace_times


async def get_booked_for_job_id(job_id: str) -> Dict:
    """
    Return the booking row for the given job_id.
    """
    response = await backend_client.get(f"/lm/api/v1/booking/job/{job_id}")
    return response.json()


def get_greatest_grace_time(job_id: str, grace_times: Dict[int, int], booking_rows: List) -> int:
    """
    Find the greatest grace_time for the given job_id.
    """
    greatest_grace_time = -1
    for book in booking_rows:
        if not book:
            continue
        for inner_book in book:
            if str(inner_book["job_id"]) != str(job_id):
                continue
            config_id_for_grace_times = inner_book["config_id"]
            greatest_grace_time = max(greatest_grace_time, grace_times[config_id_for_grace_times])
    return greatest_grace_time


def get_running_jobs(squeue_result: List) -> List:
    return [j for j in squeue_result if j["state"] == "RUNNING"]


async def clean_booked_grace_time():
    """
    Clean the booked licenses if the job's running time is greater than the grace_time.
    """
    logger.debug("GRACE_TIME START")
    formatted_squeue_output = return_formatted_squeue_out()
    cluster_name = await get_cluster_name()
    if not formatted_squeue_output:
        logger.debug("GRACE_TIME no squeue")
        await clean_bookings(None, cluster_name)
        logger.debug("GRACE_TIME cleaned bookings that are not in the queue")
        return
    squeue_result = squeue_parser(formatted_squeue_output)
    squeue_running_jobs = get_running_jobs(squeue_result)

    grace_times = await get_all_grace_times()
    get_booked_call = []
    for job in squeue_running_jobs:
        job_id = job["job_id"]
        get_booked_call.append(get_booked_for_job_id(job_id))

    booking_rows_for_running_jobs = await asyncio.gather(*get_booked_call)

    # get the greatest grace_time for each job
    for job in squeue_running_jobs:
        job_id = job["job_id"]
        greatest_grace_time = get_greatest_grace_time(job_id, grace_times, booking_rows_for_running_jobs)
        # if the running_time is greater than the greatest grace_time, delete the booking for it
        if job["run_time_in_seconds"] > greatest_grace_time and greatest_grace_time != -1:
            logger.debug(f"GRACE_TIME: {greatest_grace_time}, {job_id}")
            await remove_booked_for_job_id(job_id)
    await clean_bookings(squeue_result, cluster_name)


async def clean_bookings(squeue_result, cluster_name):
    logger.debug("CLEAN_BOOKINGS: start")
    cluster_bookings = [str(booking.job_id) for booking in await get_bookings_from_backend(cluster_name)]
    if squeue_result is None:
        squeue_result = []
    jobs_not_running = [str(job["job_id"]) for job in squeue_result if job["state"] != "RUNNING"]
    all_jobs_squeue = [str(job["job_id"]) for job in squeue_result]
    delete_booking_call = []
    logger.debug("CLEAN_BOOKINGS: after building lists")
    for job_id in cluster_bookings:
        if job_id in jobs_not_running or job_id not in all_jobs_squeue:
            delete_booking_call.append(remove_booked_for_job_id(job_id))
    logger.debug(f"CLEAN_BOOKINGS: {cluster_bookings}, {jobs_not_running}, {all_jobs_squeue}")
    if not delete_booking_call:
        logger.debug("CLEAN_BOOKINGS: no need to clean")
        return
    await asyncio.gather(*delete_booking_call)


async def filter_cluster_update_licenses(licenses_to_update: List) -> List:
    """Get the licenses in the cluster to filter the cluster update response."""
    local_licenses = await get_all_product_features_from_cluster()

    filtered_licenses = []
    for license in licenses_to_update:
        if license["product_feature"] in local_licenses:
            filtered_licenses.append(license)
    return filtered_licenses


async def reconcile():
    """Generate the report and reconcile the license feature token usage."""
    await clean_booked_grace_time()
    r = await update_report()
    response = await backend_client.get("/lm/api/v1/license/cluster_update")
    configs = await get_config_from_backend()
    licenses_to_update = response.json()

    # Filter licenses to reconcile only licenses in the cluster
    filtered_licenses = await filter_cluster_update_licenses(licenses_to_update)

    for license_data in filtered_licenses:
        product_feature = license_data["product_feature"]
        bookings_sum = license_data["bookings_sum"]
        license_total = license_data["license_total"]
        license_used = license_data["license_used"]
        config_id = await get_config_id_from_backend(product_feature)
        minimum_value = 0
        server_type = ""
        (_, feature) = product_feature.split(".")
        for config in configs:
            if config.id == config_id:
                # Using the total amount of licenses as the minimum value
                try:
                    minimum_value = config.features[feature].get("total")
                    LicenseManagerFeatureConfigurationIncorrect.require_condition(
                        minimum_value,
                        " ".join(
                            [
                                f"The configuration for {feature} is incorrect.",
                                "Please include the total amount of licenses.",
                            ]
                        ),
                    )
                except AttributeError:
                    # Fallback to get the total from the old feature format
                    minimum_value = config.features[feature]
                server_type = config.license_server_type
                break
        slurm_used = await get_tokens_for_license(product_feature + "@" + server_type, "Used")
        if slurm_used is None:
            slurm_used = 0
        new_quantity = license_total - license_used - bookings_sum + slurm_used
        if new_quantity > license_total:
            new_quantity = license_total
        if new_quantity < minimum_value:
            new_quantity = minimum_value
        product, feature = product_feature.split(".")
        update_resource = await sacctmgr_modify_resource(product, feature, new_quantity)
        if update_resource:
            logger.info("Slurmdbd updated successfully.")
        else:
            logger.info("Slurmdbd update unsuccessful.")
    return r


async def update_report():
    logger.info("Beginning forced reconciliation process")
    rep = await report()
    if not rep:
        logger.error(
            "No license data could be collected, check that tools are installed "
            "correctly and the right hosts/ports are configured in settings"
        )
        raise LicenseManagerEmptyReportError("Got an empty response from the license server")
    client = backend_client
    try:
        r = await client.patch(RECONCILE_URL_PATH, json=rep)
    except ConnectError as e:
        logger.error(f"{client.base_url}{RECONCILE_URL_PATH}: {e}")
        raise LicenseManagerBackendConnectionError("Failed to connect to the backend")

    if r.status_code != 200:
        logger.error(f"{r.url}: {r.status_code}!: {r.text}")
        raise LicenseManagerBackendConnectionError(f"Unexpected status code from report: {r.status_code}")

    logger.info(f"Forced reconciliation succeeded. backend updated: {len(rep)} feature(s)")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lm_agent',
 'lm_agent.parsing',
 'lm_agent.server_interfaces',
 'lm_agent.workload_managers',
 'lm_agent.workload_managers.slurm']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.4.0,<3.0.0',
 'boto3>=1.18.17,<2.0.0',
 'httpx>=0.18.2,<0.19.0',
 'py-buzz>=2.1.3,<3.0.0',
 'pydantic[dotenv]>=1.8.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'sentry-sdk>=1.3.1,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['reconcile = lm_agent.reconcile:main',
                     'slurmctld-epilog = '
                     'lm_agent.workload_managers.slurm.slurmctld_epilog:main',
                     'slurmctld-prolog = '
                     'lm_agent.workload_managers.slurm.slurmctld_prolog:main']}

setup_kwargs = {
    'name': 'license-manager-agent',
    'version': '2.2.21',
    'description': 'Provides an agent for interacting with license manager',
    'long_description': '# License-manager Agent',
    'author': 'OmniVector Solutions',
    'author_email': 'info@omnivector.solutions',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/omnivector-solutions/license-manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



"""
Parser for rlm
"""
import re

INT = r"\d+"
VERSION = rf"v{INT}\.{INT}"
HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTWORD2 = r"[a-zA-Z0-9-.]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD2})*"

TOTALS_LINE = rf"^(?P<license_feature>\w+) {VERSION}: (?P<user_name>\w+)@(?P<lead_host>{HOSTNAME}) (?P<license_used_by_host>\d+).*$"  # noqa
DATA_LINE = rf"^\s*(?P<license_feature>\w+) {VERSION}$"
COUNT_LINE = r"^\s*count: (?P<count>\d+).*inuse: (?P<in_use>\d+).*$"

RX_TOTAL = re.compile(TOTALS_LINE)
RX_DATA = re.compile(DATA_LINE)
RX_COUNT = re.compile(COUNT_LINE)


def _get_start_offset(lines) -> int:
    """Get the start offset of the license data."""
    i = len(lines) - 1
    count = 0
    for line in lines[::-1]:
        if "-" * 10 in line:
            count += 1
        if count == 2:
            break
        i -= 1
    return max(i, 0)


def parse(s: str) -> dict:
    """
    Parse lines of the license output with regular expressions
    """
    parsed_dict: dict = {"uses": [], "total": []}
    lines = s.splitlines()
    start_offset = _get_start_offset(lines)
    lines = [line.strip() for line in lines[start_offset:]]

    for i, line in enumerate(lines):
        parsed_total = RX_TOTAL.match(line)
        parsed_data = RX_DATA.match(line)
        if parsed_total:
            total_data = parsed_total.groupdict()
            parsed_dict["uses"].append(
                {
                    "user_name": total_data["user_name"],
                    "lead_host": total_data["lead_host"],
                    "booked": int(total_data["license_used_by_host"]),
                    "feature": total_data["license_feature"].lower(),
                }
            )
        elif parsed_data:
            parsed_count = RX_COUNT.match(lines[i + 1])
            if not parsed_count:
                continue
            count_data = parsed_count.groupdict()
            data_data = parsed_data.groupdict()
            parsed_dict["total"].append(
                {
                    "feature": data_data["license_feature"].lower(),
                    "total": int(count_data["count"]),
                    "used": int(count_data["in_use"]),
                }
            )

    return parsed_dict
